from unittest.mock import Mock

import pytest

from toga.sources import Node, TreeSource


@pytest.fixture
def listener():
    return Mock()


@pytest.fixture
def source(listener):
    source = TreeSource(
        data={
            ("group1", 1): [
                (
                    {"val1": "A first", "val2": 110},
                    None,
                ),
                (
                    {"val1": "A second", "val2": 120},
                    [],
                ),
                (
                    {"val1": "A third", "val2": 130},
                    [
                        ({"val1": "A third-first", "val2": 131}, None),
                        ({"val1": "A third-second", "val2": 132}, None),
                    ],
                ),
            ],
            ("group2", 2): [
                (
                    {"val1": "B first", "val2": 210},
                    None,
                ),
                (
                    {"val1": "B second", "val2": 220},
                    [],
                ),
                (
                    {"val1": "B third", "val2": 230},
                    [
                        ({"val1": "B third-first", "val2": 231}, None),
                        ({"val1": "B third-second", "val2": 232}, None),
                    ],
                ),
            ],
        },
        accessors=["val1", "val2"],
    )

    source.add_listener(listener)
    return source


@pytest.mark.parametrize(
    "value",
    [
        None,
        42,
        "not a list",
    ],
)
def test_invalid_accessors(value):
    """Accessors for a list source must be a list of attribute names."""
    with pytest.raises(
        ValueError,
        match=r"accessors should be a list of attribute names",
    ):
        TreeSource(accessors=value)


def test_accessors_required():
    """A list source must specify *some* accessors."""
    with pytest.raises(
        ValueError,
        match=r"TreeSource must be provided a list of accessors",
    ):
        TreeSource(accessors=[], data=[1, 2, 3])


def test_accessors_copied():
    """A list source must specify *some* accessors."""
    accessors = ["foo", "bar"]
    source = TreeSource(accessors)

    assert source._accessors == ["foo", "bar"]

    # The accessors have been copied.
    accessors.append("whiz")
    assert source._accessors == ["foo", "bar"]


@pytest.mark.parametrize(
    "data",
    [
        {},
        [],
    ],
)
def test_create_empty(data):
    """An empty TreeSource can be created."""
    source = TreeSource(data=data, accessors=["val1", "val2"])

    assert len(source) == 0


@pytest.mark.parametrize(
    "data, all_accessor_levels",
    [
        # Dictionaries all the way down
        (
            {
                "root0": {
                    "child00": None,
                    "child01": {},
                    "child02": {"child020": None, "child021": None},
                },
                "root1": {
                    "child10": None,
                    "child11": {},
                    "child12": {"child120": None, "child121": None},
                },
            },
            set(),  # Only the first accessor is ever used
        ),
        # Dictionaries with tuples as keys
        (
            {
                ("root0", 1): {
                    ("child00", 11): None,
                    ("child01", 12): {},
                    ("child02", 13): {("child020", 131): None, ("child021", 132): None},
                },
                ("root1", 2): {
                    ("child10", 21): None,
                    ("child11", 22): {},
                    ("child12", 23): {("child120", 231): None, ("child121", 232): None},
                },
            },
            {0, 1, 2},  # All accessors at all levels
        ),
        # List of dictionary data, list children
        (
            [
                (
                    {"val1": "root0", "val2": 1},
                    [
                        ({"val1": "child00", "val2": 11}, None),
                        ({"val1": "child01", "val2": 12}, []),
                        (
                            {"val1": "child02", "val2": 13},
                            [
                                ({"val1": "child020", "val2": 131}, None),
                                ({"val1": "child021", "val2": 132}, None),
                            ],
                        ),
                    ],
                ),
                (
                    {"val1": "root1", "val2": 2},
                    [
                        ({"val1": "child10", "val2": 21}, None),
                        ({"val1": "child11", "val2": 22}, []),
                        (
                            {"val1": "child12", "val2": 23},
                            [
                                ({"val1": "child120", "val2": 231}, None),
                                ({"val1": "child121", "val2": 232}, None),
                            ],
                        ),
                    ],
                ),
            ],
            {0, 1, 2},  # all accessors at all levels.
        ),
        # List of tuple data, list children
        (
            [
                (
                    ("root0", 1),
                    [
                        (("child00", 11), None),
                        (("child01", 12), []),
                        (
                            ("child02", 13),
                            [
                                (("child020", 131), None),
                                (("child021", 132), None),
                            ],
                        ),
                    ],
                ),
                (
                    ("root1", 2),
                    [
                        (("child10", 21), None),
                        (("child11", 22), []),
                        (
                            ("child12", 23),
                            [
                                (("child120", 231), None),
                                (("child121", 232), None),
                            ],
                        ),
                    ],
                ),
            ],
            {0, 1, 2},  # all accessors at all levels.
        ),
        # Dictionary of lists of dictionary data, list children
        (
            {
                "root0": [
                    ({"val1": "child00", "val2": 11}, None),
                    ({"val1": "child01", "val2": 12}, []),
                    (
                        {"val1": "child02", "val2": 13},
                        [
                            ({"val1": "child020", "val2": 131}, None),
                            ({"val1": "child021", "val2": 132}, None),
                        ],
                    ),
                ],
                "root1": [
                    ({"val1": "child10", "val2": 21}, None),
                    ({"val1": "child11", "val2": 22}, []),
                    (
                        {"val1": "child12", "val2": 23},
                        [
                            ({"val1": "child120", "val2": 231}, None),
                            ({"val1": "child121", "val2": 232}, None),
                        ],
                    ),
                ],
            },
            {1, 2},  # Accessors everywhere except the root.
        ),
        # List of dictionary data, dictionary children at level 1
        (
            [
                (
                    {"val1": "root0", "val2": 1},
                    {
                        "child00": None,
                        "child01": {},
                        "child02": [
                            ({"val1": "child020", "val2": 131}, None),
                            ({"val1": "child021", "val2": 132}, None),
                        ],
                    },
                ),
                (
                    {"val1": "root1", "val2": 2},
                    {
                        "child10": None,
                        "child11": {},
                        "child12": [
                            ({"val1": "child120", "val2": 231}, None),
                            ({"val1": "child121", "val2": 232}, None),
                        ],
                    },
                ),
            ],
            {0, 2},  # all accessors at first and last level
        ),
    ],
)
def test_create(data, all_accessor_levels):
    """A tree source can be created from data in different formats."""
    source = TreeSource(data=data, accessors=["val1", "val2"])

    # Source has 2 roots
    assert len(source) == 2

    # Root0 has 2 children
    assert source[0].val1 == "root0"
    assert len(source[0]) == 3
    assert source[0].can_have_children()

    # Root1 has 2 children
    assert source[1].val1 == "root1"
    assert len(source[1]) == 3
    assert source[1].can_have_children()

    # If level 0 has all accessors, check them as well.
    if 0 in all_accessor_levels:
        assert source[0].val2 == 1
        assert source[1].val2 == 2

    # Children of root 0
    assert source[0][0].val1 == "child00"
    assert len(source[0][0]) == 0
    assert not source[0][0].can_have_children()

    assert source[0][1].val1 == "child01"
    assert len(source[0][1]) == 0
    assert source[0][1].can_have_children()

    assert source[0][2].val1 == "child02"
    assert len(source[0][2]) == 2
    assert source[0][2].can_have_children()

    # Children of root 1
    assert source[1][0].val1 == "child10"
    assert len(source[1][0]) == 0
    assert not source[1][0].can_have_children()

    assert source[1][1].val1 == "child11"
    assert len(source[1][1]) == 0
    assert source[1][1].can_have_children()

    assert source[1][2].val1 == "child12"
    assert len(source[1][2]) == 2
    assert source[1][2].can_have_children()

    # If level 1 has all accessors, check them as well.
    if 1 in all_accessor_levels:
        assert source[0][0].val2 == 11
        assert source[0][1].val2 == 12
        assert source[0][2].val2 == 13

        assert source[1][0].val2 == 21
        assert source[1][1].val2 == 22
        assert source[1][2].val2 == 23

    # Children of root 0, child 2
    assert source[0][2][0].val1 == "child020"
    assert len(source[0][2][0]) == 0
    assert not source[0][2][0].can_have_children()

    assert source[0][2][1].val1 == "child021"
    assert len(source[0][2][1]) == 0
    assert not source[0][2][1].can_have_children()

    # Children of root 1, child 2
    assert source[1][2][0].val1 == "child120"
    assert len(source[1][2][0]) == 0
    assert not source[1][2][0].can_have_children()

    assert source[1][2][1].val1 == "child121"
    assert len(source[1][2][1]) == 0
    assert not source[1][2][1].can_have_children()

    # If level 2 has all accessors, check them as well.
    if 2 in all_accessor_levels:
        assert source[0][2][0].val2 == 131
        assert source[0][2][1].val2 == 132

        assert source[1][2][0].val2 == 231
        assert source[1][2][1].val2 == 232


def test_source_single_object():
    """A single object can be passed as root data."""
    source = TreeSource(accessors=["val1", "val2"], data="A string")

    assert len(source) == 1
    assert source[0].val1 == "A string"


def test_single_object_child():
    """A single object can be passed as child data."""
    source = TreeSource(
        accessors=["val1", "val2"],
        data={("root1", 1): "A string"},
    )

    assert len(source) == 1
    assert source[0].val1 == "root1"
    assert source[0].val2 == 1
    assert source[0].can_have_children()

    assert len(source[0]) == 1
    assert len(source[0][0]) == 0
    assert source[0][0].val1 == "A string"
    assert not source[0][0].can_have_children()


def test_modify_roots(source, listener):
    """The roots of a source can be modified."""
    root = source[1]
    assert root.val1 == "group2"

    # delete the root
    del source[1]

    # Removal notification was sent
    listener.remove.assert_called_once_with(parent=None, index=1, item=root)
    listener.reset_mock()

    # Root is no longer associated with the source
    assert root._parent is None
    assert root._source is None

    # Root count has dropped
    assert len(source) == 1

    old_root_0 = source[0]

    # A child can be modified by index
    source[0] = {"val1": "new"}

    # Root 0 has changed instance
    assert old_root_0 is not source[0]

    # Old child 0 is no longer associated with this node
    assert old_root_0._source is None
    assert old_root_0._parent is None

    # Change notification was sent, the change is associated with the new item
    listener.change.assert_called_once_with(item=source[0])

    # Source's root count hasn't changed
    assert len(source) == 1


def test_iter_root(source):
    """The roots of a source can be iterated over."""
    assert "|".join(root.val1 for root in source) == "group1|group2"


def test_clear(source, listener):
    """A TreeSource can be cleared."""
    source.clear()

    assert len(source) == 0

    # Clear notification was sent
    listener.clear.assert_called_once_with()


@pytest.mark.parametrize(
    "index, actual_index",
    [
        (1, 1),  # Positive, in range
        (10, 2),  # Positive, past positive limit
        (-1, 1),  # Negative, in range
        (-10, 0),  # Negative, past negative limit
    ],
)
def test_insert(source, listener, index, actual_index):
    """A new root node can be inserted."""
    new_child = source.insert(index, {"val1": "new"})

    # Source has one more root.
    assert len(source) == 3
    assert source[actual_index] == new_child

    # Root data is as expected
    assert source[actual_index].val1 == "new"

    # Insert notification was sent, the change is associated with the new item
    listener.insert.assert_called_once_with(
        parent=None,
        index=actual_index,
        item=new_child,
    )


def test_insert_with_children(source, listener):
    """A new root node can be inserted with children."""
    new_child = source.insert(
        1,
        {"val1": "new"},
        children=[
            ({"val1": "new child 1"}, None),
            ({"val1": "new child 2"}, None),
        ],
    )

    # Source has one more root.
    assert len(source) == 3
    assert source[1] == new_child

    # Root data is as expected
    assert source[1].val1 == "new"
    assert len(source[1]) == 2

    # Children are also present
    assert source[1][0].val1 == "new child 1"
    assert not source[1][0].can_have_children()
    assert source[1][1].val1 == "new child 2"
    assert not source[1][1].can_have_children()

    # Insert notification was sent, the change is associated with the new item
    listener.insert.assert_called_once_with(
        parent=None,
        index=1,
        item=new_child,
    )


def test_append(source, listener):
    """A new root node can be appended."""
    new_child = source.append({"val1": "new"})

    # Source has one more root.
    assert len(source) == 3
    assert source[2] == new_child

    # Root data is as expected
    assert source[2].val1 == "new"

    # Insert notification was sent, the change is associated with the new item
    listener.insert.assert_called_once_with(
        parent=None,
        index=2,
        item=new_child,
    )


def test_append_with_children(source, listener):
    """A new root node can be inserted with children."""
    new_child = source.append(
        {"val1": "new"},
        children=[
            ({"val1": "new child 1"}, None),
            ({"val1": "new child 2"}, None),
        ],
    )

    # Source has one more root.
    assert len(source) == 3
    assert source[2] == new_child

    # Root data is as expected
    assert source[2].val1 == "new"
    assert len(source[2]) == 2

    # Children are also present
    assert source[2][0].val1 == "new child 1"
    assert not source[2][0].can_have_children()
    assert source[2][1].val1 == "new child 2"
    assert not source[2][1].can_have_children()

    # Insert notification was sent, the change is associated with the new item
    listener.insert.assert_called_once_with(
        parent=None,
        index=2,
        item=new_child,
    )


def test_remove_root(source, listener):
    """A root node can be removed."""
    root = source[1]
    source.remove(root)

    # One less item in the source
    assert len(source) == 1

    # The root is no longer associated with the source
    assert root._source is None

    # Removal notification was sent
    listener.remove.assert_called_once_with(parent=None, index=1, item=root)


def test_remove_child(source, listener):
    """A child node can be removed from a source."""
    node = source[1][1]
    source.remove(node)

    # The source still has 2 roots
    assert len(source) == 2
    # ... but there's 1 less child
    assert len(source[1]) == 2

    # The child is no longer associated with the source,
    # and the child isn't associated with its parent.
    assert node._source is None
    assert node._parent is None

    # Removal notification was sent
    listener.remove.assert_called_once_with(parent=source[1], index=1, item=node)


def test_remove_non_root(source, listener):
    """If a node isn't associated with this source, remove raises an error."""
    other = Node(val="other")

    with pytest.raises(
        ValueError,
        match=r"<Leaf Node .*> is not managed by this data source",
    ):
        source.remove(other)


def test_index(source):
    """A root can be found in a TreeSource."""
    root = source[1]
    assert source.index(root) == 1


def test_find(source):
    """A node can be found by value."""
    root1 = source[1]

    # Append some additional roots
    root2 = source.append({"val1": "group1", "val2": 333})
    root3 = source.append({"val1": "group2", "val2": 444})

    # Find the child by a partial match of values.
    assert source.find({"val1": "group2"}) == root1

    # Find the child by a partial match of values, starting at the first match
    assert source.find({"val1": "group2"}, start=root1) == root3

    # Find the child by a full match of values, starting at the first match
    assert source.find({"val1": "group1", "val2": 333}) == root2



from pathlib import Path

import pytest

import toga


class MyDoc(toga.Document):
    def __init__(self, path, app):
        super().__init__(path, "Dummy Document", app)
        pass

    def create(self):
        pass

    def read(self):
        pass


@pytest.mark.parametrize("path", ["/path/to/doc.mydoc", Path("/path/to/doc.mydoc")])
def test_create_document(app, path):
    doc = MyDoc(path, app)

    assert doc.path == Path(path)
    assert doc.app == app
    assert doc.document_type == "Dummy Document"


class MyDeprecatedDoc(toga.Document):
    def __init__(self, filename, app):
        super().__init__(
            path=filename,
            document_type="Deprecated Document",
            app=app,
        )

    def create(self):
        pass

    def read(self):
        pass


def test_deprecated_names(app):
    """Deprecated names still work."""
    doc = MyDeprecatedDoc("/path/to/doc.mydoc", app)

    with pytest.warns(
        DeprecationWarning,
        match=r"Document.filename has been renamed Document.path.",
    ):
        assert doc.filename == Path("/path/to/doc.mydoc")



from pathlib import Path
from unittest.mock import ANY

import PIL.Image
import pytest
from travertino.size import at_least

import toga
from toga.style.pack import Pack
from toga.widgets.imageview import rehint_imageview
from toga_dummy.utils import (
    assert_action_not_performed,
    assert_action_performed,
    assert_action_performed_with,
)


@pytest.fixture
def widget(app):
    return toga.ImageView()


def test_create_empty(widget):
    """An empty ImageView can be created."""
    # interface/impl round trips
    assert widget._impl.interface is widget
    assert_action_performed(widget, "create ImageView")
    assert_action_performed_with(widget, "set image", image=None)
    assert_action_performed(widget, "refresh")

    assert widget.image is None


ABSOLUTE_FILE_PATH = Path(__file__).parent.parent / "resources/toga.png"


def test_create_from_toga_image(app):
    """An ImageView can be created from a Toga image."""
    image = toga.Image(ABSOLUTE_FILE_PATH)
    widget = toga.ImageView(image=image)

    # Interface/impl round trips
    assert widget._impl.interface is widget
    assert_action_performed(widget, "create ImageView")
    assert_action_performed_with(widget, "set image", image=image)
    assert_action_performed(widget, "refresh")

    # Image attribute is set
    assert widget.image == image


def test_create_from_pil():
    """An ImageView can be created from a PIL image."""
    with PIL.Image.open(ABSOLUTE_FILE_PATH) as pil_img:
        pil_img.load()

    imageview = toga.ImageView(pil_img)
    assert isinstance(imageview.image, toga.Image)
    assert imageview.image.size == (32, 32)


def test_disable_no_op(widget):
    """ImageView doesn't have a disabled state."""

    # Enabled by default
    assert widget.enabled

    # Try to disable the widget
    widget.enabled = False

    # Still enabled.
    assert widget.enabled


def test_focus_noop(widget):
    """Focus is a no-op."""
    widget.focus()
    assert_action_not_performed(widget, "focus")


def test_set_image_str(widget):
    """The image can be set with a string."""
    widget.image = ABSOLUTE_FILE_PATH

    assert_action_performed_with(widget, "set image", image=ANY)
    assert_action_performed(widget, "refresh")

    assert isinstance(widget.image, toga.Image)
    assert widget.image.path == ABSOLUTE_FILE_PATH


def test_set_image_path(widget):
    """The image can be set with a Path."""
    widget.image = Path(ABSOLUTE_FILE_PATH)

    assert_action_performed_with(widget, "set image", image=ANY)
    assert_action_performed(widget, "refresh")

    assert isinstance(widget.image, toga.Image)
    assert widget.image.path == ABSOLUTE_FILE_PATH


def test_set_image(widget):
    """The image can be set with an Image instance."""
    image = toga.Image(Path(ABSOLUTE_FILE_PATH))
    widget.image = image
    assert_action_performed_with(widget, "set image", image=image)
    assert_action_performed(widget, "refresh")

    assert widget.image == image


def test_set_image_none(app):
    """The image can be cleared."""
    widget = toga.ImageView(image=ABSOLUTE_FILE_PATH)
    assert widget.image is not None

    widget.image = None
    assert_action_performed_with(widget, "set image", image=None)
    assert_action_performed(widget, "refresh")

    assert widget.image is None


@pytest.mark.parametrize(
    "params, expected_width, expected_height, expected_aspect_ratio",
    [
        # Intrinsic image size
        (dict(style=Pack()), 144, 72, 2),
        (dict(style=Pack(), scale=2), 288, 144, 2),
        # Fixed width
        (dict(style=Pack(width=150)), 150, 75, 2),
        (dict(style=Pack(width=150), scale=2), 300, 150, 2),
        # Fixed height
        (dict(style=Pack(height=80)), 160, 80, 2),
        (dict(style=Pack(height=80), scale=2), 320, 160, 2),
        # Explicit image size
        (dict(style=Pack(width=37, height=42)), 37, 42, None),
        (dict(style=Pack(width=37, height=42), scale=2), 74, 84, None),
        # Intrinsic image size, flex widget
        (dict(style=Pack(flex=1)), at_least(0), at_least(0), 2),
        (dict(style=Pack(flex=1), scale=2), at_least(0), at_least(0), 2),
        # Fixed width, flex widget
        (dict(style=Pack(width=150, flex=1)), 150, at_least(0), 2),
        (dict(style=Pack(width=150, flex=1), scale=2), 300, at_least(0), 2),
        # Fixed height, flex widget
        (dict(style=Pack(height=80, flex=1)), at_least(0), 80, 2),
        (dict(style=Pack(height=80, flex=1), scale=2), at_least(0), 160, 2),
        # Explicit image size, flex widget
        (dict(style=Pack(width=37, height=42, flex=1)), 37, 42, None),
        (dict(style=Pack(width=37, height=42, flex=1), scale=2), 74, 84, None),
    ],
)
def test_rehint_image(
    app,
    params,
    expected_width,
    expected_height,
    expected_aspect_ratio,
):
    path = Path(__file__).parent.parent / "resources/sample.png"
    image = toga.Image(path)

    width, height, aspect_ratio = rehint_imageview(image=image, **params)
    assert width == expected_width
    assert height == expected_height
    assert aspect_ratio == expected_aspect_ratio


@pytest.mark.parametrize(
    "params",
    [
        # Default scale
        dict(style=Pack()),
        # Explicit width/height
        dict(style=Pack(width=100)),
        dict(style=Pack(height=200)),
        dict(style=Pack(width=100, height=200)),
        # 2x Scale
        dict(style=Pack(), scale=2),
    ],
)
def test_rehint_empty_image(params):
    width, height, aspect_ratio = rehint_imageview(image=None, **params)
    assert width == 0
    assert height == 0
    assert aspect_ratio is None


def test_as_image_toga():
    toga_img = toga.Image(ABSOLUTE_FILE_PATH)
    imageview = toga.ImageView(toga_img)
    toga_img_2 = imageview.as_image()
    assert isinstance(toga_img_2, toga.Image)
    assert toga_img_2.size == (32, 32)


def test_as_image_pil():
    imageview = toga.ImageView(ABSOLUTE_FILE_PATH)
    pil_img = imageview.as_image(PIL.Image.Image)
    assert isinstance(pil_img, PIL.Image.Image)
    assert pil_img.size == (32, 32)



from unittest.mock import Mock

import pytest

import toga
from toga.sources import ListSource
from toga_dummy.utils import (
    EventLog,
    assert_action_not_performed,
    assert_action_performed,
    assert_action_performed_with,
)


@pytest.fixture
def on_change_handler():
    return Mock()


@pytest.fixture
def source():
    return ListSource(
        accessors=["key", "value"],
        data=[
            {"key": "first", "value": 111},
            {"key": "second", "value": 222},
            {"key": "third", "value": 333},
        ],
    )


@pytest.fixture
def widget(source, on_change_handler):
    return toga.Selection(
        accessor="key",
        items=source,
        value=source[1],
        on_change=on_change_handler,
    )


def test_widget_created():
    """An empty selection can be created."""
    widget = toga.Selection()
    assert widget._impl.interface == widget
    assert_action_performed(widget, "create Selection")

    assert len(widget.items) == 0
    assert widget._accessor is None
    assert widget.value is None
    assert widget.on_change._raw is None
    assert widget.enabled


def test_create_with_value():
    """A Selection can be created with initial values."""
    on_change = Mock()

    widget = toga.Selection(
        items=["first", "second", "third"],
        value="second",
        on_change=on_change,
        enabled=False,
    )
    assert widget._impl.interface == widget
    assert_action_performed(widget, "create Selection")

    assert len(widget.items) == 3
    assert widget._accessor is None
    assert widget.value == "second"
    assert widget.on_change._raw == on_change
    assert not widget.enabled


@pytest.mark.parametrize(
    "items, title, value",
    [
        # list of strings
        (["first", "second", "third"], "second", "second"),
        # list of non-strings
        ([111, 222, 333], "222", 222),
        # List of dictionaries; implied use of "value" key
        (
            [
                {"key": "first", "value": 111},
                {"key": "second", "value": 222},
                {"key": "third", "value": 333},
            ],
            "222",
            222,
        ),
        # List of tuples; only the first item is used.
        (
            [
                ("first", 111),
                ("second", 222),
                ("third", 333),
            ],
            "second",
            "second",
        ),
        # List of strings with a newline in value
        (["first", "second\nitem", "third"], "second", "second\nitem"),
        # List of strings with a duplicate entry
        (["first", "second", "third", "first", "second"], "second", "second"),
    ],
)
def test_value_no_accessor(items, title, value):
    """If there's no accessor, the items can be set and values will be dereferenced."""
    on_change_handler = Mock()
    widget = toga.Selection(on_change=on_change_handler)

    # There haven't been any changes
    on_change_handler.assert_not_called()

    # Change the items. This changes the selection, triggering the handler
    widget.items = items

    on_change_handler.assert_called_once_with(widget)

    # Default selection is the first item
    # As there's no accessor, the value auto-dereferences
    assert widget.value == widget.items[0].value

    # Reset the mock
    on_change_handler.reset_mock()

    # Change the value
    widget.value = value
    assert widget.value == value

    # The value renders as a string
    assert widget._title_for_item(widget.items[1]) == title

    # Change handler was invoked
    on_change_handler.assert_called_once_with(widget)


@pytest.mark.parametrize(
    "accessor, items, title, value",
    [
        # List of strings with a custom accessor name
        ("key", ["first", "second", "third"], "second", "second"),
        # List of dictionaries selecting non-default key
        (
            "key",
            [
                {"key": "first", "value": 111},
                {"key": "second", "value": 222},
                {"key": "third", "value": 333},
            ],
            "second",
            "second",
        ),
        # List of tuples; accessor name is altered.
        (
            "key",
            [
                ("first", 111),
                ("second", 222),
                ("third", 333),
            ],
            "second",
            "second",
        ),
        (
            "key",
            ListSource(
                accessors=["key", "value"],
                data=[
                    {"key": "first", "value": 111},
                    {"key": "second", "value": 222},
                    {"key": "third", "value": 333},
                ],
            ),
            "second",
            "second",
        ),
        # List of strings with a newline in a title
        ("key", ["first", "second\nitem", "third"], "second", "second\nitem"),
        # List of strings with a duplicate entry
        ("key", ["first", "second", "third", "first", "second"], "second", "second"),
    ],
)
def test_value_with_accessor(accessor, items, title, value):
    """If an accessor is used, item lookup semantics are different."""
    on_change_handler = Mock()
    widget = toga.Selection(accessor=accessor, on_change=on_change_handler)

    # There haven't been any changes
    on_change_handler.assert_not_called()

    # Change the items. This changes the selection, triggering the handler
    widget.items = items

    on_change_handler.assert_called_once_with(widget)

    # Default selection is the first item
    # As there is an accessor, we get back a literal row object.
    assert widget.value == widget.items[0]

    # Reset the mock
    on_change_handler.reset_mock()

    # Change the value. As there is an accessor, we use a literal row object,
    # and need to manually access attributes on that row.
    widget.value = widget.items[1]
    assert getattr(widget.value, accessor) == value

    # The value renders as a string
    assert widget._title_for_item(widget.items[1]) == title

    # Change handler was invoked
    on_change_handler.assert_called_once_with(widget)


def test_source_no_accessor():
    """If the selection is based on a source, it must specify an accessor."""
    with pytest.raises(
        ValueError,
        match=r"Must specify an accessor to use a data source",
    ):
        toga.Selection(items=ListSource(accessors=["foo", "bar"]))


def test_bad_item_no_accessor():
    """The value for the selection must exist in accessor-less data."""
    selection = toga.Selection(items=["first", "second", "third"])

    with pytest.raises(
        ValueError,
        match=r"'bad' is not a current item in the selection",
    ):
        selection.value = "bad"


def test_bad_item_with_accessor():
    """The value for the selection must exist in accessor-based data."""
    # Create a selection with an extra item and an explicit accessor
    selection = toga.Selection(
        accessor="value", items=["first", "bad", "second", "third"]
    )

    # Store the bad item
    item = selection.items.find(dict(value="bad"))

    # Delete the bad item
    selection.items.remove(item)

    # item is no longer a valid selection
    with pytest.raises(
        ValueError,
        match=r"<Row .* value='bad'> is not a current item in the selection",
    ):
        selection.value = item


def test_add_item(widget, source, on_change_handler):
    """A row can be added to the source."""
    # Store the original selection
    selection = widget.value

    source.append(dict(key="new", value=999))

    # The widget adds the item
    assert_action_performed_with(widget, "insert item", index=3)

    # This doesn't change the widget.
    on_change_handler.assert_not_called()
    assert widget.value == selection


def test_insert_item(widget, source, on_change_handler):
    """A row can be inserted into the source."""
    # Store the original selection
    selection = widget.value

    source.insert(1, dict(key="new", value=999))

    # The widget adds the item
    assert_action_performed_with(widget, "insert item", index=1)

    # This doesn't change the widget
    on_change_handler.assert_not_called()
    assert widget.value == selection


def test_remove(widget, source, on_change_handler):
    """If you remove an item that isn't selected, no change is generated."""
    # Store the original selection
    selection = widget.value

    # Remove a non-selected item
    item = source[0]
    source.remove(item)

    # The widget adds the item
    assert_action_performed_with(widget, "remove item", index=0, item=item)

    # This changes the selection
    on_change_handler.assert_not_called()
    assert widget.value == selection


def test_remove_selected(widget, source, on_change_handler):
    """If you remove the currently selected item, a change is generated."""
    # Store the original selection
    selection = widget.value

    source.remove(selection)

    # The widget adds the item
    assert_action_performed_with(widget, "remove item", index=1, item=selection)

    # This changes the selection
    on_change_handler.assert_called_with(widget)
    assert widget.value == source[0]


def test_clear_source(widget, source, on_change_handler):
    """If the source is cleared, the selection is cleared."""
    # Clear the source
    source.clear()

    # The widget has been cleared
    assert_action_performed(widget, "clear")

    # The widget must have cleared its selection
    on_change_handler.assert_called_with(widget)
    assert widget.value is None


def test_change_source_empty(widget, on_change_handler):
    """If the source is changed to an empty source, the selection is reset."""
    # Clear the event history
    EventLog.reset()

    widget.items = []

    # The widget data has been cleared and refreshed
    assert_action_performed(widget, "clear")
    assert_action_not_performed(widget, "insert item")
    assert_action_performed(widget, "refresh")

    # The widget must have cleared its selection
    on_change_handler.assert_called_once_with(widget)
    assert widget.value is None


def test_change_source(widget, on_change_handler):
    """If the source is changed, the selection is set to the first item."""
    # Clear the event history
    EventLog.reset()

    # Change the source of the data
    widget.items = ["new 1", "new 2"]

    # The widget source has changed
    assert_action_performed(widget, "clear")
    assert_action_performed_with(widget, "insert item", item=widget.items[0])
    assert_action_performed_with(widget, "insert item", item=widget.items[1])
    assert_action_performed(widget, "refresh")

    # The widget must have cleared its selection
    on_change_handler.assert_called_once_with(widget)
    assert widget.value.key == "new 1"


######################################################################
# 2023-05: Backwards compatibility
######################################################################


def test_deprecated_names(on_change_handler):
    """Deprecated names still work."""

    # Can't specify both on_select and on_change
    with pytest.raises(
        ValueError,
        match=r"Cannot specify both on_select and on_change",
    ):
        toga.Selection(on_select=Mock(), on_change=Mock())

    # on_select is redirected at construction
    with pytest.warns(
        DeprecationWarning,
        match="Selection.on_select has been renamed Selection.on_change",
    ):
        select = toga.Selection(on_select=on_change_handler)

    # on_select accessor is redirected to on_change
    with pytest.warns(
        DeprecationWarning,
        match="Selection.on_select has been renamed Selection.on_change",
    ):
        assert select.on_select._raw == on_change_handler

    assert select.on_change._raw == on_change_handler

    # on_select mutator is redirected to on_change
    new_handler = Mock()
    with pytest.warns(
        DeprecationWarning,
        match="Selection.on_select has been renamed Selection.on_change",
    ):
        select.on_select = new_handler

    assert select.on_change._raw == new_handler



from __future__ import annotations

from typing import TYPE_CHECKING

from travertino.size import at_least

import toga
from toga.style.pack import NONE
from toga.widgets.base import Widget

if TYPE_CHECKING:
    from toga.images import ImageContent, ImageT


def rehint_imageview(image, style, scale=1):
    """Compute the size hints for an ImageView based on the image.

    This logic is common across all backends, so it's shared here.

    :param image: The image being displayed.
    :param style: The style object for the imageview.
    :param scale: The scale factor (if any) to apply to native pixel sizes.
    :returns: A triple containing the intrinsic width hint, intrinsic height hint, and
        the aspect ratio to preserve (or None if the aspect ratio should not be
        preserved).
    """
    if image:
        if style.width != NONE and style.height != NONE:
            # Explicit width and height for image. Scale the rendered image
            # to fit the explicitly provided size.
            width = int(style.width * scale)
            height = int(style.height * scale)
            aspect_ratio = None

        elif style.width != NONE:
            # Explicit width, implicit height. Preserve aspect ratio.
            aspect_ratio = image.width / image.height
            width = int(style.width * scale)
            height = int(style.width * scale / aspect_ratio)
            if style.flex:
                height = at_least(0)
        elif style.height != NONE:
            # Explicit height, implicit width. Preserve aspect ratio.
            aspect_ratio = image.width / image.height
            width = int(style.height * scale * aspect_ratio)
            height = int(style.height * scale)
            if style.flex:
                width = at_least(0)
        else:
            # Use the image's actual size.
            aspect_ratio = image.width / image.height
            width = int(image.width * scale)
            height = int(image.height * scale)
            if style.flex:
                width = at_least(0)
                height = at_least(0)
    else:
        # No image. Hinted size is 0.
        width = 0
        height = 0
        aspect_ratio = None

    return width, height, aspect_ratio


# Note: remove PIL type annotation when plugin system is implemented for image format
# registration; replace with ImageT?
class ImageView(Widget):
    def __init__(
        self,
        image: ImageContent | None = None,
        id=None,
        style=None,
    ):
        """
        Create a new image view.

        :param image: The image to display. Can be any valid :any:`image content
            <ImageContent>` type; or :any:`None` to display no image.
        :param id: The ID for the widget.
        :param style: A style object. If no style is provided, a default style will be
            applied to the widget.
        """
        super().__init__(id=id, style=style)
        # Prime the image attribute
        self._image = None
        self._impl = self.factory.ImageView(interface=self)
        self.image = image

    @property
    def enabled(self) -> bool:
        """Is the widget currently enabled? i.e., can the user interact with the widget?

        ImageView widgets cannot be disabled; this property will always return True; any
        attempt to modify it will be ignored.
        """
        return True

    @enabled.setter
    def enabled(self, value):
        pass

    def focus(self):
        "No-op; ImageView cannot accept input focus"
        pass

    @property
    def image(self) -> toga.Image | None:
        """The image to display.

        When setting an image, you can provide any valid :any:`image content
        <ImageContent>` type; or :any:`None` to clear the image view.
        """
        return self._image

    @image.setter
    def image(self, image):
        if isinstance(image, toga.Image):
            self._image = image
        elif image is None:
            self._image = None
        else:
            self._image = toga.Image(image)

        self._impl.set_image(self._image)
        self.refresh()

    def as_image(self, format: type[ImageT] = toga.Image) -> ImageT:
        """Return the image in the specified format.

        :param format: Format to provide. Defaults to :class:`~toga.images.Image`; also
            supports :any:`PIL.Image.Image` if Pillow is installed, as well as any image
            types defined by installed :doc:`image format plugins
            </reference/plugins/image_formats>`.
        :returns: The image in the specified format.
        """
        return self.image.as_format(format)
